# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
IBM XForce Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, ClassVar

from typing_extensions import Self

from ..._version import VERSION
from ...common.utility import export
from ..http_provider import APILookupParams
from .ti_http_provider import HttpTIProvider
from .ti_provider_base import ResultSeverity

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-few-public-methods
@dataclass
class _XForceParams(APILookupParams):
    # override APILookupParams to set common defaults
    def __attrs_post_init__(self: Self) -> None:
        self.auth_str = ["{ApiID}", "{AuthKey}"]
        self.auth_type = "HTTPBasic"


@export
class XForce(HttpTIProvider):
    """IBM XForce Lookup."""

    _BASE_URL: ClassVar[str] = "https://api.xforce.ibmcloud.com"

    _QUERIES: ClassVar[dict[str, _XForceParams]] = {
        "ipv4": _XForceParams(path="/ipr/{observable}"),
        "ipv4-rep": _XForceParams(path="/ipr/history/{observable}"),
        "ipv4-malware": _XForceParams(path="/ipr/malware/{observable}"),
        "ipv4-whois": _XForceParams(path="/whois/{observable}"),
        "dns-passivedns": _XForceParams(path="/resolve/{observable}"),
        "hostname-whois": _XForceParams(path="/whois/{observable}"),
        "file_hash": _XForceParams(path="/malware/{observable}"),
        "url": _XForceParams(path="/url/{observable}"),
        "url-malware": _XForceParams(path="/url/malware/{observable}"),
    }

    # aliases
    _QUERIES["ipv6"] = _QUERIES["ipv4"]
    _QUERIES["ipv6-rep"] = _QUERIES["ipv4-rep"]
    _QUERIES["ipv6-malware"] = _QUERIES["ipv4-malware"]
    _QUERIES["ipv6-whois"] = _QUERIES["ipv4-whois"]
    _QUERIES["md5_hash"] = _QUERIES["file_hash"]
    _QUERIES["sha1_hash"] = _QUERIES["file_hash"]
    _QUERIES["sha256_hash"] = _QUERIES["file_hash"]
    _QUERIES["dns"] = _QUERIES["url"]
    _QUERIES["dns-malware"] = _QUERIES["url-malware"]
    _QUERIES["ipv4-passivedns"] = _QUERIES["dns-passivedns"]
    _QUERIES["ipv6-passivedns"] = _QUERIES["dns-passivedns"]
    _QUERIES["hostname-whois"] = _QUERIES["ipv4-whois"]
    _QUERIES["dns-whois"] = _QUERIES["ipv4-whois"]

    _REQUIRED_PARAMS: ClassVar[list[str]] = ["ApiID", "AuthKey"]

    HIGH_SEVERITY: ClassVar[int] = 5
    MEDIUM_SEVERITY: ClassVar[int] = 2

    def parse_results(self: Self, response: dict) -> tuple[bool, ResultSeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : Dict
            The returned data response

        Returns
        -------
        Tuple[bool, ResultSeverity, Any]
            bool = positive or negative hit
            ResultSeverity = enumeration of severity
            Object with match details

        """
        severity: ResultSeverity = ResultSeverity.information
        if self._failed_response(response) or not isinstance(
            response["RawResult"],
            dict,
        ):
            return False, severity, "Not found."
        result = True
        result_dict: dict[str, Any] = {}
        if (
            response["IocType"] in ["ipv4", "ipv6", "url", "dns"]
            and not response["QuerySubtype"]
        ):
            # For some IocTypes "dns" and "url", the response structure differs
            report: dict[str, Any] = response["RawResult"].get(
                "result",
                response["RawResult"],
            )
            score: int = report.get("score", 0)
            if score is None:
                score = 0
            result_dict.update(
                {
                    "score": report.get("score", 0),
                    "cats": report.get("cats"),
                    "categoryDescriptions": report.get("categoryDescriptions"),
                    "reason": report.get("reason"),
                    "reasonDescription": report.get("reasonDescription", 0),
                    "tags": report.get("tags", 0),
                },
            )
            severity = (
                ResultSeverity.information
                if score < self.MEDIUM_SEVERITY
                else (
                    ResultSeverity.warning
                    if self.MEDIUM_SEVERITY <= score < self.HIGH_SEVERITY
                    else ResultSeverity.high
                )
            )
        if (
            response["IocType"] in ["file_hash", "md5_hash", "sha1_hash", "sha256_hash"]
            or response["QuerySubtype"] == "malware"
        ):
            malware: dict[str, Any] | None = response["RawResult"].get("malware")
            if malware:
                result_dict.update(
                    {
                        "risk": malware.get("risk"),
                        "family": malware.get("family"),
                        "reasonDescription": response["RawResult"].get(
                            "reasonDescription",
                            0,
                        ),
                    },
                )
                severity = ResultSeverity.high
        if response["IocType"] in ["dns", "ipv4", "ipv6", "hostname"] and response[
            "QuerySubtype"
        ] in ["info", "passivedns", "whois"]:
            records: int = response["RawResult"].get("total_rows", 0)
            contact: int = response["RawResult"].get("contact", 0)
            if records:
                result_dict["records"] = records
            elif contact:
                result_dict["contact"] = contact
        return result, severity, result_dict



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""MDATP OData Driver class."""
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union

import pandas as pd

from ..._version import VERSION
from ...auth.azure_auth_core import AzureCloudConfig
from ...auth.cloud_mappings import (
    get_defender_endpoint,
    get_m365d_endpoint,
    get_m365d_login_endpoint,
)
from ...common.data_utils import ensure_df_datetimes
from ...common.utility import export
from ..core.query_defns import DataEnvironment
from .odata_driver import OData, QuerySource, _get_driver_settings

__version__ = VERSION
__author__ = "Pete Bryan"


@dataclass
class M365DConfiguration:
    """A container for M365D API settings.

    This is based on the data environment of the query provider.
    """

    login_uri: str
    resource_uri: str
    api_version: str
    api_endpoint: str
    api_uri: str
    scopes: List[str]
    oauth_v2: bool = field(init=False)

    def __post_init__(self):
        """Determine if the selected API supports Entra ID OAuth v2.0.

        This is important because the fields in the request body
        are different between the two versions.
        """
        if "/oauth2/v2.0" in self.login_uri:
            self.oauth_v2 = True
        else:
            self.oauth_v2 = False


@export
class MDATPDriver(OData):
    """KqlDriver class to retrieve date from MS Defender APIs."""

    CONFIG_NAME = "MicrosoftDefender"
    _ALT_CONFIG_NAMES = ["MDATPApp"]

    def __init__(
        self, connection_str: Optional[str] = None, instance: str = "Default", **kwargs
    ):
        """
        Instantiate MSDefenderDriver and optionally connect.

        Parameters
        ----------
        connection_str : str, optional
            Connection string
        instance : str, optional
            The instance name from config to use

        """
        super().__init__(**kwargs)

        cs_dict = _get_driver_settings(
            self.CONFIG_NAME, self._ALT_CONFIG_NAMES, instance
        )

        self.cloud = cs_dict.pop("cloud", "global")
        if "cloud" in kwargs and kwargs["cloud"]:
            self.cloud = kwargs["cloud"]

        m365d_params = _select_api(self.data_environment, self.cloud)
        self._m365d_params: M365DConfiguration = m365d_params
        self.oauth_url = m365d_params.login_uri
        self.api_root = m365d_params.resource_uri
        self.api_ver = m365d_params.api_version
        self.api_suffix = m365d_params.api_endpoint
        self.scopes = m365d_params.scopes

        self.add_query_filter(
            "data_environments", ("MDE", "M365D", "MDATP", "M365DGraph", "GraphHunting")
        )

        self.req_body: Dict[str, Any] = {}
        if "username" in cs_dict:
            delegated_auth = True

        else:
            delegated_auth = False
            self.req_body["grant_type"] = "client_credentials"

        if not m365d_params.oauth_v2:
            self.req_body["resource"] = self.scopes

        if connection_str:
            self.current_connection = connection_str
            self.connect(
                connection_str,
                delegated_auth=delegated_auth,
                auth_type=kwargs.get("auth_type", "interactive"),
                location=cs_dict.get("location", "token_cache.bin"),
            )

    def query(
        self, query: str, query_source: Optional[QuerySource] = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Returns
        -------
        Union[pd.DataFrame, results.ResultSet]
            A DataFrame (if successful) or
            the underlying provider result if an error.

        """
        del query_source, kwargs
        data, response = self.query_with_results(
            query, body=True, api_end=self.api_suffix
        )
        if isinstance(data, pd.DataFrame):
            # If we got a schema we should convert the DateTimes to pandas datetimes
            if ("Schema" or "schema") not in response:
                return data

            if self.data_environment == DataEnvironment.M365DGraph:
                date_fields = [
                    field["name"]
                    for field in response["schema"]
                    if field["type"] == "DateTime"
                ]
            else:
                date_fields = [
                    field["Name"]
                    for field in response["Schema"]
                    if field["Type"] == "DateTime"
                ]
            data = ensure_df_datetimes(data, columns=date_fields)
            return data
        return response


def _select_api(data_environment, cloud) -> M365DConfiguration:
    # pylint: disable=line-too-long
    """Return API and login URIs for selected provider type.

    Note that the Microsoft Graph is the preferred API.

    | API Name | Resource ID | Scopes Requested | API URI (global cloud) | API Endpoint | Login URI | MSTICpy Data Environment |
    | -------- | ----------- | ---------------- | ---------------------- | ------------ | --------- | ------------------------ |
    | WindowsDefenderATP | fc780465-2017-40d4-a0c5-307022471b92 | `AdvancedQuery.Read` | `https://api.securitycenter.microsoft.com` | `/advancedqueries/run` | `https://login.microsoftonline.com/<tenantId>/oauth2/token` | `MDE`, `MDATP` |
    | Microsoft Threat Protection | 8ee8fdad-f234-4243-8f3b-15c294843740 | `AdvancedHunting.Read` | `https://api.security.microsoft.com` | `/advancedhunting/run` | `https://login.microsoftonline.com/<tenantId>/oauth2/token` | `M365D` |
    | Microsoft Graph | 00000003-0000-0000-c000-000000000000 | `ThreatHunting.Read.All` | `https://graph.microsoft.com/<version>/` | `/security/runHuntingQuery` | `https://login.microsoftonline.com/<tenantId>/oauth2/v2.0/token` | `M365DGraph` |

    """
    # pylint: enable=line-too-long
    if data_environment == DataEnvironment.M365DGraph:
        az_cloud_config = AzureCloudConfig(cloud=cloud)
        login_uri = f"{az_cloud_config.authority_uri}{{tenantId}}/oauth2/v2.0/token"
        resource_uri = az_cloud_config.endpoints["microsoftGraphResourceId"]
        api_version = "v1.0"
        api_endpoint = "/security/runHuntingQuery"
        scopes = [f"{resource_uri}ThreatHunting.Read.All"]

    elif data_environment == DataEnvironment.M365D:
        login_uri = f"{get_m365d_login_endpoint(cloud)}{{tenantId}}/oauth2/token"
        resource_uri = get_m365d_endpoint(cloud)
        api_version = "api"
        api_endpoint = "/advancedhunting/run"
        scopes = [f"{resource_uri}AdvancedHunting.Read"]

    else:
        login_uri = f"{get_m365d_login_endpoint(cloud)}{{tenantId}}/oauth2/token"
        resource_uri = get_defender_endpoint(cloud)
        api_version = "api"
        api_endpoint = "/advancedqueries/run"
        scopes = [f"{resource_uri}AdvancedQuery.Read"]

    api_uri = f"{resource_uri}{api_version}{api_endpoint}"

    return M365DConfiguration(
        login_uri=login_uri,
        resource_uri=resource_uri,
        api_version=api_version,
        api_endpoint=api_endpoint,
        api_uri=api_uri,
        scopes=scopes,
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Code cleanup functions to re-format obfuscated code."""
import re

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


def format_powershell(code: str) -> str:
    """
    Cleanup obfuscated powershell.

    Parameters
    ----------
    code : str
        Powershell code segment.

    Returns
    -------
    str
       Reformatted code.

    Notes
    -----
    This function is purely for displaying intentionally obfuscated
    code, it will likely not return executable code.

    See Also
    --------
    msticpy.vis.code_view

    """
    # pre-process: change to lowercase and remove redundant string concat
    cleaned_str = code.casefold().replace("'+'", "")

    # Use regex for more complex replacements
    # regex patterns to replace
    patterns = {
        ";": "\n",
        "{": "\n{\n",
        "}": "\n}\n",
        "`": "",
    }
    # replace only if not in a quoted string
    search_pattern = r"(\"|')[^\"']+\"'|(\{find})"

    # iterate through patterns replacing each one.
    for pattern, repl in patterns.items():
        cleaned_str = re.sub(search_pattern.format(find=pattern), repl, cleaned_str)

    # re-process lines
    cleaned_lines = cleaned_str.splitlines()
    indent = 0
    output = []
    for line in cleaned_lines:
        # add indents if line starts with "{" and de-indent on "}"
        if not line:
            continue  # skip blank lines
        if line.startswith("}"):
            indent -= 4
        output.append(" " * indent + line)
        if line.startswith("{"):
            indent += 4

    return "\n".join(output)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Query hierarchy attribute class."""
from functools import partial

from ..._version import VERSION
from ...common.data_types import ObjectContainer

__version__ = VERSION
__author__ = "Ian Hellen"


class QueryContainer(ObjectContainer):
    """Empty class used to create hierarchical attributes."""

    def __repr__(self):
        """Return list of attributes."""
        repr_list = []
        for name, obj in self.__dict__.items():
            if isinstance(obj, QueryContainer):
                repr_list.append(f"{name} (container)")
            elif isinstance(obj, partial):
                repr_list.append(f"{name} (query)")
            elif not name.startswith("_"):
                repr_list.append(f"{name} {type(obj).__name__}")
        return "\n".join(repr_list)

    def __call__(self, *args, **kwargs):
        """Return list of attributes or help."""
        if args or kwargs:
            print("This attribute is a container, not a query.")
            print("Items in this container:")
        print(repr(self))
